/*
 *	LAPB release 002
 *
 *	This code REQUIRES 2.1.15 or higher/ NET3.038
 *
 *	This module:
 *		This module is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 */

#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_defines.h>
#include <linux/wanpipe_debug.h>
#include <linux/wanpipe_common.h>
#include <linux/wanpipe_cfg.h>
#include <linux/wanpipe_lapb.h>

/*
 *	State machine for state 0, Disconnected State.
 *	The handling of the timer(s) is in file lapb_timer.c.
 */
static void lapb_state0_machine(lapb_cb *lapb, struct sk_buff *skb, struct lapb_frame *frame)
{
	switch (frame->type) {
		case LAPB_SABM:

			if (test_bit(LAPB_DOWN,&lapb->critical)){
				return;
			}
			
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S0 RX SABM(%d)\n", lapb->dev, frame->pf);
#endif
			if (lapb->mode & LAPB_EXTENDED) {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S0 TX DM(%d)\n", lapb->dev, frame->pf);
#endif
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			} else {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S0 TX UA(%d)\n", lapb->dev, frame->pf);
#endif
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S0 -> S3\n", lapb->dev);
#endif
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_CONNECTED);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_connect_indication(lapb, LAPB_OK);
			}
			break;

		case LAPB_SABME:

			if (test_bit(LAPB_DOWN,&lapb->critical)){
				return;
			}

#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S0 RX SABME(%d)\n", lapb->dev, frame->pf);
#endif
			if (lapb->mode & LAPB_EXTENDED) {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S0 TX UA(%d)\n", lapb->dev, frame->pf);
#endif
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S0 -> S3\n", lapb->dev);
#endif
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_CONNECTED);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_connect_indication(lapb, LAPB_OK);
			} else {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S0 TX DM(%d)\n", lapb->dev, frame->pf);
#endif
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			}
			break;

		case LAPB_DISC:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S0 RX DISC(%d)\n", lapb->dev, frame->pf);
			printk(KERN_DEBUG "lapb: (%p) S0 TX UA(%d)\n", lapb->dev, frame->pf);
#endif
			lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
			break;

		default:
			break;
	}

	KFREE_SKB(skb);
}

/*
 *	State machine for state 1, Awaiting Connection State.
 *	The handling of the timer(s) is in file lapb_timer.c.
 */
static void lapb_state1_machine(lapb_cb *lapb, struct sk_buff *skb, struct lapb_frame *frame)
{
	switch (frame->type) {
		case LAPB_SABM:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S1 RX SABM(%d)\n", lapb->dev, frame->pf);
#endif
			if (lapb->mode & LAPB_EXTENDED) {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S1 TX DM(%d)\n", lapb->dev, frame->pf);
#endif
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			} else {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S1 TX UA(%d)\n", lapb->dev, frame->pf);
#endif
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				
				if (frame->pf) {

					lapb_stop_t1timer(lapb);
					lapb_stop_t2timer(lapb);
					lapb_state_change(lapb,LAPB_STATE_CONNECTED);
					lapb->condition = 0x00;
					lapb->n2count   = 0;
					lapb->vs        = 0;
					lapb->vr        = 0;
					lapb->va        = 0;
					lapb_connect_confirmation(lapb, LAPB_OK);
				}
			}
			break;

		case LAPB_SABME:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S1 RX SABME(%d)\n", lapb->dev, frame->pf);
#endif
			if (lapb->mode & LAPB_EXTENDED) {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S1 TX UA(%d)\n", lapb->dev, frame->pf);
#endif
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				
				if (frame->pf) {

					lapb_stop_t1timer(lapb);
					lapb_stop_t2timer(lapb);
					lapb_state_change(lapb,LAPB_STATE_CONNECTED);
					lapb->condition = 0x00;
					lapb->n2count   = 0;
					lapb->vs        = 0;
					lapb->vr        = 0;
					lapb->va        = 0;
					lapb_connect_confirmation(lapb, LAPB_OK);
				}
			} else {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S1 TX DM(%d)\n", lapb->dev, frame->pf);
#endif
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			}
			break;

		case LAPB_DISC:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S1 RX DISC(%d)\n", lapb->dev, frame->pf);
			printk(KERN_DEBUG "lapb: (%p) S1 TX DM(%d)\n", lapb->dev, frame->pf);
#endif
			lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			break;

		case LAPB_UA:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S1 RX UA(%d)\n", lapb->dev, frame->pf);
#endif
			if (frame->pf) {
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S1 -> S3\n", lapb->dev);
#endif
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_CONNECTED);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_connect_confirmation(lapb, LAPB_OK);
			}
			break;

		case LAPB_DM:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S1 RX DM(%d)\n", lapb->dev, frame->pf);
#endif
			if (frame->pf) {
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S1 -> S0\n", lapb->dev);
#endif
				lapb_clear_queues(lapb);
				lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_disconnect_indication(lapb, LAPB_REFUSED);
			}
			break;

		default:
			break;
	}

	KFREE_SKB(skb);
}

/*
 *	State machine for state 2, Awaiting Release State.
 *	The handling of the timer(s) is in file lapb_timer.c
 */
static void lapb_state2_machine(lapb_cb *lapb, struct sk_buff *skb, struct lapb_frame *frame)
{
	switch (frame->type) {
		case LAPB_SABM:
		case LAPB_SABME:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S2 RX {SABM,SABME}(%d)\n", lapb->dev, frame->pf);
			printk(KERN_DEBUG "lapb: (%p) S2 TX DM(%d)\n", lapb->dev, frame->pf);
#endif
			lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			break;

		case LAPB_DISC:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S2 RX DISC(%d)\n", lapb->dev, frame->pf);
			printk(KERN_DEBUG "lapb: (%p) S2 TX UA(%d)\n", lapb->dev, frame->pf);
#endif
			lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
			break;

		case LAPB_UA:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S2 RX UA(%d)\n", lapb->dev, frame->pf);
#endif
			if (frame->pf) {
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S2 -> S0\n", lapb->dev);
#endif
				lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_disconnect_confirmation(lapb, LAPB_OK);
			}
			break;

		case LAPB_DM:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S2 RX DM(%d)\n", lapb->dev, frame->pf);
#endif
			if (frame->pf) {
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S2 -> S0\n", lapb->dev);
#endif
				lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_disconnect_confirmation(lapb, LAPB_NOTCONNECTED);
			}
			break;

		case LAPB_I:
		case LAPB_REJ:
		case LAPB_RNR:
		case LAPB_RR:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S2 RX {I,REJ,RNR,RR}(%d)\n", lapb->dev, frame->pf);
			printk(KERN_DEBUG "lapb: (%p) S2 RX DM(%d)\n", lapb->dev, frame->pf);
#endif
			if (frame->pf) lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			break;
				
		default:
			break;
	}

	KFREE_SKB(skb);
}

/*
 *	State machine for state 3, Connected State.
 *	The handling of the timer(s) is in file lapb_timer.c
 */
static void lapb_state3_machine(lapb_cb *lapb, struct sk_buff *skb, struct lapb_frame *frame)
{
	int queued = 0;
	int modulus;
	
	modulus = (lapb->mode & LAPB_EXTENDED) ? LAPB_EMODULUS : LAPB_SMODULUS;

	lapb->nr_type=frame->type;

	switch (frame->type) {
		case LAPB_SABM:
			
			printk(KERN_INFO "%s: LAPB Event : State Connected : Rx SABM !\n",lapb->dev->name);
			
			if (lapb->mode & LAPB_EXTENDED) {
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			} else {
				lapb_clear_queues(lapb);
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_requeue_frames(lapb);
			}

			lapb_connect_indication(lapb, LAPB_OK);
			break;

		case LAPB_SABME:
			
			printk(KERN_INFO "%s: LAPB Event : State Connected : Rx SABME !\n",lapb->dev->name);

			if (lapb->mode & LAPB_EXTENDED) {
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_requeue_frames(lapb);
			} else {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S3 TX DM(%d)\n", lapb->dev, frame->pf);
#endif
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			}
			break;

		case LAPB_DISC:
			printk(KERN_INFO "%s: LAPB Event : State connected : Rx DISC\n",lapb->dev->name);	
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S3 RX DISC(%d)\n", lapb->dev, frame->pf);
#endif
#if LAPB_DEBUG > 0
			printk(KERN_DEBUG "lapb: (%p) S3 -> S0\n", lapb->dev);
#endif
			lapb_clear_queues(lapb);
			lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
			lapb_start_t1timer(lapb);
			lapb_stop_t2timer(lapb);
			lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
			lapb_disconnect_indication(lapb, LAPB_OK);
			break;

		case LAPB_DM:
			printk(KERN_INFO "%s: LAPB Event : State connected : Rx DM\n",lapb->dev->name);
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S3 RX DM(%d)\n", lapb->dev, frame->pf);
#endif
#if LAPB_DEBUG > 0
			printk(KERN_DEBUG "lapb: (%p) S3 -> S0\n", lapb->dev);
#endif
			lapb_clear_queues(lapb);
			lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
			lapb_start_t1timer(lapb);
			lapb_stop_t2timer(lapb);
			lapb_disconnect_indication(lapb, LAPB_NOTCONNECTED);
			break;

		case LAPB_RNR:

#if LAPB_DEBUG > 1
			printk(KERN_INFO "%s: LAPB Event : State connected : Rx RNR\n",lapb->dev->name);
			printk(KERN_DEBUG "lapb: (%p) S3 RX RNR(%d) R%d\n", lapb->dev, frame->pf, frame->nr);
#endif
			lapb->condition |= LAPB_PEER_RX_BUSY_CONDITION;
			lapb_check_need_response(lapb, frame->cr, frame->pf);

			if (lapb_validate_nr(lapb, frame->nr)) {
				lapb_check_iframes_acked(lapb, frame->nr);
			} else {
				lapb->old_nr_type=frame->type;

				lapb->frmr_data = *frame;
				lapb->frmr_type = LAPB_FRMR_Z;
				lapb_transmit_frmr(lapb);
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S3 -> S4\n", lapb->dev);
#endif
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_FRAME_REJECT);
				lapb->n2count = 0;
			}
			break;

		case LAPB_RR:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S3 RX RR(%d) R%d\n", lapb->dev, frame->pf, frame->nr);
#endif
			lapb->condition &= ~LAPB_PEER_RX_BUSY_CONDITION;
			lapb_check_need_response(lapb, frame->cr, frame->pf);
			if (lapb_validate_nr(lapb, frame->nr)) {
				lapb_check_iframes_acked(lapb, frame->nr);
			} else {
				lapb->old_nr_type=frame->type;

				lapb->frmr_data = *frame;
				lapb->frmr_type = LAPB_FRMR_Z;
				lapb_transmit_frmr(lapb);
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S3 -> S4\n", lapb->dev);
#endif
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_FRAME_REJECT);
				lapb->n2count = 0;
			}
			break;

		case LAPB_REJ:
			
#if LAPB_DEBUG > 1
			printk(KERN_INFO "%s: LAPB Event : State Connected : Rx REJ\n",lapb->dev->name);
			printk(KERN_DEBUG "lapb: (%p) S3 RX REJ(%d) R%d\n", lapb->dev, frame->pf, frame->nr);
#endif
			lapb->condition &= ~LAPB_PEER_RX_BUSY_CONDITION;
			lapb_check_need_response(lapb, frame->cr, frame->pf);

			if (lapb_validate_nr(lapb, frame->nr)) {

				lapb_frames_acked(lapb, frame->nr);
				lapb_stop_t1timer(lapb);
				lapb->n2count = 0;
				lapb_requeue_frames(lapb);

			} else {
				lapb->old_nr_type=frame->type;

				lapb->frmr_data = *frame;
				lapb->frmr_type = LAPB_FRMR_Z;
				lapb_transmit_frmr(lapb);
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S3 -> S4\n", lapb->dev);
#endif
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_FRAME_REJECT);
				lapb->n2count = 0;
			}
			break;

		case LAPB_I:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S3 RX I(%d) S%d R%d\n", lapb->dev, frame->pf, frame->ns, frame->nr);
#endif
			if (!lapb_validate_nr(lapb, frame->nr)) {

				lapb->old_nr_type=frame->type;

				printk(KERN_INFO "%s: Lapb Error: I-Frame Rx invalid NR count %i !!!\n",
						lapb->dev->name,frame->nr);
				
				lapb->frmr_data = *frame;
				lapb->frmr_type = LAPB_FRMR_Z;
				lapb_transmit_frmr(lapb);
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S3 -> S4\n", lapb->dev);
#endif
				lapb_start_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_FRAME_REJECT);
				lapb->n2count = 0;
				break;
			}

			if (lapb->condition & LAPB_PEER_RX_BUSY_CONDITION){
				lapb_frames_acked(lapb, frame->nr);
			}else{
				lapb_check_iframes_acked(lapb, frame->nr);
			}
			
			if (frame->ns == lapb->vr){

				struct sk_buff *newskb;
				
				/*
				 * If upper layer has dropped the frame, we
				 * basically ignore any further protocol
				 * processing. This will cause the peer
				 * to re-transmit the frame later like
				 * a frame lost on the wire.
				 */

				queued = 1;	

				skb_queue_tail(&lapb->rx_queue,skb);
				lapb->vr = (lapb->vr + 1) % modulus;

				while ((newskb=skb_dequeue(&lapb->rx_queue)) != NULL){
					if (lapb_data_indication(lapb, newskb) != 0){
						printk(KERN_INFO "%s: Failed to pass data up : top layer busy!\n",
								lapb->dev->name);
						skb_queue_head(&lapb->rx_queue,newskb);
						break;
					}
				}
				
				lapb->condition &= ~LAPB_REJECT_CONDITION;
				
				if (frame->pf){
					lapb_enquiry_response(lapb);
				}else{
					if (!(lapb->condition & LAPB_ACK_PENDING_CONDITION)){
					
						lapb->condition |= LAPB_ACK_PENDING_CONDITION;

						/* If the rx queue has reached maximum, send
						 * out an RNR to flow control the incoming
						 * traffic, untill the queue is offloaded by
						 * the above protocol */
						if (skb_queue_len(&lapb->rx_queue) > LAPB_RX_Q_H_LIMIT){
							lapb->condition |= LAPB_RNR_SENT_CONDITION; 
							DEBUG_TEST("%s: LAPB receive queue limit! Sending RNR!\n",
									lapb->dev->name);
							lapb_send_control(lapb, LAPB_RNR, 
							        LAPB_POLLOFF, LAPB_RESPONSE);
							break;
						}

						/* If the T2 timer is zero, then send the
						 * RR response to the I frame right away,
						 * otherwise start the T2 timer and send
						 * the response after timer expires 
						 *
						 * NC: T2==0 -> send RR after every frame!
						 *     However, don't send RR if there
						 *     are I-frames pending.  RR is an
						 *     overkill.
						 * */
						if (!lapb->t2){
							if (!skb_queue_len(&lapb->write_queue) &&
							    !skb_queue_len(&lapb->async_queue)){
								lapb_send_control(lapb, LAPB_RR, 
									  LAPB_POLLOFF, LAPB_RESPONSE);
							}else{
								DEBUG_TEST("Lapb Q WQ=%i AQ=%i\n",
										skb_queue_len(&lapb->write_queue),
										skb_queue_len(&lapb->async_queue));
							}
							lapb->condition &= ~LAPB_ACK_PENDING_CONDITION;
						}else{
							lapb_start_t2timer(lapb);
						}
					}
				}
			} else {
				if (lapb->condition & LAPB_REJECT_CONDITION) {
					if (frame->pf)
						lapb_enquiry_response(lapb);
				} else {
#if LAPB_DEBUG > 1
					printk(KERN_DEBUG "lapb: (%p) S3 TX REJ(%d) R%d\n", lapb->dev, frame->pf, lapb->vr);
#endif
					DEBUG_TEST("%s: Error: Lapb sending reject: In Ns=%i != Orig Vr=%i\n",
							lapb->dev->name,frame->ns,lapb->vr);

					lapb->condition |= LAPB_REJECT_CONDITION;
					lapb_send_control(lapb, LAPB_REJ, frame->pf, LAPB_RESPONSE);
					lapb->condition &= ~LAPB_ACK_PENDING_CONDITION;
				}
			}
			break;

		case LAPB_FRMR:

			printk(KERN_INFO "%s: LAPB Event : State Connected : Rx FRMR\n",lapb->dev->name);
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S3 RX FRMR(%d) %02X %02X %02X %02X %02X\n", lapb->dev, frame->pf, skb->data[0], skb->data[1], skb->data[2], skb->data[3], skb->data[4]);
#endif
			lapb_establish_data_link(lapb);
#if LAPB_DEBUG > 0
			printk(KERN_DEBUG "lapb: (%p) S3 -> S1\n", lapb->dev);
#endif
	
			lapb_requeue_frames(lapb);
			lapb_state_change(lapb,LAPB_STATE_CONNECTING);
			break;

		case LAPB_ILLEGAL:

#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S3 RX ILLEGAL(%d)\n", lapb->dev, frame->pf);
#endif
			lapb->frmr_data = *frame;
			lapb->frmr_type = LAPB_FRMR_W;
			lapb_transmit_frmr(lapb);
#if LAPB_DEBUG > 0
			printk(KERN_DEBUG "lapb: (%p) S3 -> S4\n", lapb->dev);
#endif
			lapb_start_t1timer(lapb);
			lapb_stop_t2timer(lapb);
			lapb_state_change(lapb,LAPB_STATE_FRAME_REJECT);
			lapb->n2count = 0;
			break;

		default:
			break;
	}

	if (!queued){
		KFREE_SKB(skb);
	}
}

/*
 *	State machine for state 4, Frame Reject State.
 *	The handling of the timer(s) is in file lapb_timer.c.
 */
static void lapb_state4_machine(lapb_cb *lapb, struct sk_buff *skb, struct lapb_frame *frame)
{
	switch (frame->type) {
		case LAPB_SABM:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S4 RX SABM(%d)\n", lapb->dev, frame->pf);
#endif
			if (lapb->mode & LAPB_EXTENDED) {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S4 TX DM(%d)\n", lapb->dev, frame->pf);
#endif
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			} else {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S4 TX UA(%d)\n", lapb->dev, frame->pf);
#endif
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S4 -> S3\n", lapb->dev);
#endif
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_CONNECTED);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_connect_indication(lapb, LAPB_OK);
			}
			break;

		case LAPB_SABME:
#if LAPB_DEBUG > 1
			printk(KERN_DEBUG "lapb: (%p) S4 RX SABME(%d)\n", lapb->dev, frame->pf);
#endif
			if (lapb->mode & LAPB_EXTENDED) {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S4 TX UA(%d)\n", lapb->dev, frame->pf);
#endif
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S4 -> S3\n", lapb->dev);
#endif
				lapb_send_control(lapb, LAPB_UA, frame->pf, LAPB_RESPONSE);
				lapb_stop_t1timer(lapb);
				lapb_stop_t2timer(lapb);
				lapb_state_change(lapb,LAPB_STATE_CONNECTED);
				lapb->condition = 0x00;
				lapb->n2count   = 0;
				lapb->vs        = 0;
				lapb->vr        = 0;
				lapb->va        = 0;
				lapb_connect_indication(lapb, LAPB_OK);
			} else {
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S4 TX DM(%d)\n", lapb->dev, frame->pf);
#endif
				lapb_send_control(lapb, LAPB_DM, frame->pf, LAPB_RESPONSE);
			}
			break;

		default:
			break;
	}

	KFREE_SKB(skb);
}

/*
 *	Process an incoming LAPB frame
 */
void lapb_data_input(lapb_cb *lapb, struct sk_buff *skb)
{
	struct lapb_frame frame;
	int len = skb->len;


	if (test_bit(LAPB_UNREG_CRIT,&lapb->critical)){
		KFREE_SKB(skb);
		return;
	}

	if (!(lapb->dev->flags & IFF_UP)){
		KFREE_SKB(skb);
		return;
	}
	
	if (test_and_set_bit(LAPB_RX_SMP,&lapb->critical)){
		printk(KERN_INFO "---!!!__ Critical in Lapb Rx!\n");
	}
	
	lapb_decode(lapb, skb, &frame);

	if (frame.type == LAPB_CORRUPT){
		KFREE_SKB(skb);
		++lapb->ifstats.rx_errors;
		goto lapb_data_input_error;
	}

	if (skb->len > lapb->mtu){
		printk(KERN_INFO "LAPB RX: incoming packet size %i exceed maximum %i\n",
				skb->len, lapb->mtu);
		frame.type = LAPB_ILLEGAL;
	}
	
	lapb_update_stats(lapb,frame.type,len,0);

	/* When we receive a valid frame, restart
	 * the T4 timer. Used for link inactivity
	 * detection */
	if (frame.type != LAPB_ILLEGAL && !test_bit(LAPB_DOWN,&lapb->critical)){
		lapb_start_t4timer(lapb, 1);
	}

	//spin_lock(&lapb->bh_lock);

	switch (lapb->state) {
		case LAPB_STATE_DISCONNECTED:
			lapb_state0_machine(lapb, skb, &frame);
			break;
		case LAPB_STATE_CONNECTING:
			lapb_state1_machine(lapb, skb, &frame);
			break;
		case LAPB_STATE_DISCONNECTING:
			lapb_state2_machine(lapb, skb, &frame);
			break;
		case LAPB_STATE_CONNECTED:
			lapb_state3_machine(lapb, skb, &frame);
			break;
		case LAPB_STATE_FRAME_REJECT:
			lapb_state4_machine(lapb, skb, &frame);
			break;
		default:
			printk(KERN_INFO "%s: Lapb Rx in invalid state: droping packet\n",
					lapb->dev->name);
			KFREE_SKB(skb);
			break;
	}

	//spin_unlock(&lapb->bh_lock);

lapb_data_input_error:
	clear_bit(LAPB_RX_SMP,&lapb->critical);
	lapb_trigger_kick(lapb,1);
}

